from url_cleaner.url import Url
from url_cleaner.filter import Filter
from url_cleaner.ruler import Ruler
import os
import logging

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

MODULE_DIR = os.path.dirname(os.path.abspath(__file__))

class UrlCleaner(object):

    def __init__(self,rule_path=MODULE_DIR):
        self.ruler = Ruler(rule_path)

    def clean(self,url):
        url_parsed = Url(url)
        logger.debug('URL parsed:{}'.format(url_parsed.get_dict()))
        rules = self.ruler.get_general_rules() + self.ruler.host_rules(url_parsed.host)
        logger.debug("Rules:{}".format(rules))
        filter = Filter(rules)
        cleaned_url = filter.filter_url(url_parsed)
        return cleaned_url


import re

class Filter(object):
    def __init__(self, rules):
        self.rules = rules

    def filter_url(self, url):
        self.url = url
        for rule in self.rules:
            if isinstance(rule,str):
                self.remove_query(rule)
            elif isinstance(rule,dict):
                assert rule.keys() == {'pathname','search'}
                if self.check_pathname(rule['pathname']):
                    assert isinstance(rule['search'],str)
                    self.remove_query(rule['search'])
        return self.url.get_url()

    def remove_query(self, query):
        if query.startswith('/') and query.endswith('/'):
            self.remove_query_by_regex(query)
        elif self.url.query_dict.get(query):
            self.url.query_dict.pop(query)

    def check_pathname(self, pathname):
        if pathname.startswith('/') and pathname.endswith('/'):
            return bool(re.search(pathname[1:-1],self.url.path))
        else:
            return pathname in self.url.path

    def remove_query_by_regex(self, regex):
        regex = r"{}".format(regex[1:-1])
        cleaned_query_dict = {}
        for k in self.url.query_dict.keys():
            if bool(re.search(regex,k)):
                continue
            cleaned_query_dict[k] = self.url.query_dict[k]
        self.url.query_dict = cleaned_query_dict


from urllib.parse import urlparse, urlunparse, parse_qs, urlencode

class Url(object):
    """
    The following methods are used to modify the url
    """
    def __init__(self, url=None):
        self.scheme = None
        self.netloc = None
        self.host = None
        self.path = None
        self.params = None
        self.query = None
        self.fragment = None
        self.query_dict = None
        if url:
            self.original_url = url
            self.parse_url()
            self.parse_query()

    def parse_url(self):
        """
        Parse the url and set the attributes
        """
        if not self.original_url.startswith('http'):
            self.original_url = 'http://' + self.original_url
        u = urlparse(self.original_url)
        self.scheme = u.scheme
        self.netloc = u.netloc
        self.host = u.hostname
        self.path = u.path
        self.params = u.params
        self.query = u.query
        self.fragment = u.fragment
        return u

    def parse_query(self):
        """
        Parse the query string and set the query_dict attribute
        """
        self.query_dict = parse_qs(self.query)
        return self.query_dict

    def get_query_by_dict(self):
        """
        Return the query string by the query dict
        """
        return urlencode(self.query_dict, doseq=True)

    def get_url(self):
        """
        Unparse the url and return the url
        """
        self.query = self.get_query_by_dict()
        return urlunparse((self.scheme,self.netloc, self.path, self.params, self.query, self.fragment))

    def __str__(self):
        return self.get_url()

    def copy(self):
        """
        Return a copy of the Url object
        """
        return Url(self.get_url())

    def get_dict(self):
        """
        transform the object to dict
        """
        return {
            'scheme': self.scheme,
            'netloc': self.netloc,
            'host': self.host,
            'path': self.path,
            'params': self.params,
            'query': self.query,
            'fragment': self.fragment,
            'query_dict': self.query_dict,
        }



import json
import os
import re
import logging

from url_cleaner.url import Url

logger = logging.getLogger(__name__)

class Ruler(object):
    rules_remote_specific = "https://github.com/AdguardTeam/AdguardFilters/raw/master/TrackParamFilter/sections/specific.txt"
    rules_remote_general = "https://github.com/AdguardTeam/AdguardFilters/raw/master/TrackParamFilter/sections/general_url.txt"
    def __init__(self, module_path):
        self.rules_file = os.path.join(module_path, 'rules.json')
        self.module_path = module_path
        self.load_rules()

    def check_rules_exists(self):
        return os.path.exists(self.rules_file)
    
    def load_rules(self):
        if self.check_rules_exists():
            with open(self.rules_file) as f:
                self.rules = json.load(f)
        else:
            raise FileNotFoundError('Rules file not found')

    def save_rules(self):
        with open(self.rules_file, 'w') as f:
            json.dump(self.rules, f, indent=4,ensure_ascii=False)
    
    def download_remote_rules(self,mode=""):
        import requests
        if mode == "specific" or mode == "":
            r = requests.get(self.rules_remote_specific)
            if r.status_code == 200:
                with open(os.path.join(self.module_path, 'rules_remote_specific.txt'), 'w') as f:
                    f.write(r.text)
            else:
                raise Exception('Download remote specific rules failed:{}'.format(self.rules_remote_specific))
        if mode == "general" or mode == "":
            r = requests.get(self.rules_remote_general)
            if r.status_code == 200:
                with open(os.path.join(self.module_path, 'rules_remote_general.txt'), 'w') as f:
                    f.write(r.text)
            else:
                raise Exception('Download remote general rules failed:{}'.format(self.rules_remote_general))

    def update_rules(self,mode=""):
        add_cnt = {"specific":0,"general":0}
        if mode not in ['general', 'specific','']:
            raise Exception('Invalid mode:{}'.format(mode))
        if mode == "specific" or mode == "":
            self.download_remote_rules(mode="specific")
            with open(os.path.join(self.module_path, 'rules_remote_specific.txt')) as f:
                rules_remote_specific = f.read()
            for line in rules_remote_specific.split('\n'):
                ru = re.search(r"^\|\|([^\^]+)\^\$removeparam=([^\,\|]+)",line)
                if ru:
                    url = ru.group(1)
                    query = ru.group(2)
                    url_parsed = Url("http://"+url)
                    host = url_parsed.host
                    pathname = url_parsed.path
                    if host not in self.rules.keys():
                        self.rules[url_parsed.host] = []
                    if len(pathname) >= 1 and pathname != "/":
                        new_rule = {"pathname":pathname,"search":query}
                    else:
                        new_rule = query
                    if new_rule not in self.rules[host]:
                        self.rules[host].append(new_rule)
                        add_cnt["specific"] += 1
                        logger.info('Add specific rule:{} {}'.format(host,str(new_rule)))
        if mode == "general" or mode == "":
            self.download_remote_rules(mode="general")
            with open(os.path.join(self.module_path, 'rules_remote_general.txt')) as f:
                rules_remote_general = f.read()
            for line in rules_remote_general.split('\n'):
                ru = re.search(r'^\$removeparam=([\w\-\_]+)$',line)
                if ru and ru.group(1) not in self.rules["GENERAL"]:
                    self.rules["GENERAL"].append(ru.group(1))
                    add_cnt["general"] += 1
                    logger.info('Add general rule:{}'.format(ru.group(1)))
        logger.info('Update rules finished, add {} specific rules, add {} general rules'.format(add_cnt["specific"],add_cnt["general"]))
        self.save_rules()

    def host_rules(self,host):
        hosts_rules = []
        for rule_host in self.rules.keys():
            if re.search(rule_host,host):
                hosts_rules += self.rules[rule_host]
        return hosts_rules

    def get_general_rules(self):
        return self.rules["GENERAL"]


from url_cleaner import UrlCleaner

if __name__ == '__main__':
    c = UrlCleaner()
    
    urls = [
        'https://baijiahao.baidu.com/s?id=1748839822649920321&wfr=spider&for=pc',
        'https://mp.weixin.qq.com/s?__biz=MjM5OTExMjYwMA==&mid=2670081058&idx=6&sn=1ad7112020c2a4104d67ca542ab14444&chksm=bc12eed58b6567c30c78123a9e890ed79b448f82642305dabae4fa1f507a57f5ce0ac7a85554#rd',
        'https://www.bilibili.com/video/BV158411b7ki/?spm_id_from=333.1007.tianma.1-2-2.click',
    ]

    for url in urls:
        print("原始：",url)
        cleaned = c.clean(url)
        print("清除后：",cleaned)
        print('================================')

    # Update rules
    c.ruler.update_rules(mode="specific")
    c.ruler.update_rules(mode="general")